/*
comm_cuda.cu (CUDA + MPI)
*/

#ifdef _MPI
#include <mpi.h>
#endif

#include "ofd.h"
#include "ofd_cuda.h"
#include "ofd_prototype.h"

// share boundary H : copy D2H + comm + copy H2D
void comm_cuda_boundary(void)
{
#ifdef _MPI
	MPI_Status status;
	const int tag = 0;

	int64_t offset_hy[4], offset_hz[4], length_hy, length_hz;
	comm_buffer(offset_hy, offset_hz, &length_hy, &length_hz);

	const int count = (int)(length_hy + length_hz);
	const size_t size_hy = length_hy * sizeof(real_t);
	const size_t size_hz = length_hz * sizeof(real_t);
	real_t *sendbuf = (real_t *)malloc(size_hy + size_hz);
	real_t *recvbuf = (real_t *)malloc(size_hy + size_hz);

	// -X boundary
	if (commRank > 0) {
		// copy device to host
		cuda_memcpy(GPU, sendbuf,             Hy + offset_hy[1], size_hy, cudaMemcpyDeviceToHost);
		cuda_memcpy(GPU, sendbuf + length_hy, Hz + offset_hz[1], size_hz, cudaMemcpyDeviceToHost);

		// MPI
		const int dst = commRank - 1;
		MPI_Sendrecv(sendbuf, count, MPI_REAL_T, dst, tag,
		             recvbuf, count, MPI_REAL_T, dst, tag, MPI_COMM_WORLD, &status);

		// copy host to device
		cuda_memcpy(GPU, Hy + offset_hy[0], recvbuf,             size_hy, cudaMemcpyHostToDevice);
		cuda_memcpy(GPU, Hz + offset_hz[0], recvbuf + length_hy, size_hz, cudaMemcpyHostToDevice);
	}

	// +X boundary
	if (commRank < commSize - 1) {
		// copy device to host
		cuda_memcpy(GPU, sendbuf,             Hy + offset_hy[2], size_hy, cudaMemcpyDeviceToHost);
		cuda_memcpy(GPU, sendbuf + length_hy, Hz + offset_hz[2], size_hz, cudaMemcpyDeviceToHost);

		// MPI
		const int dst = commRank + 1;
		MPI_Sendrecv(sendbuf, count, MPI_REAL_T, dst, tag,
		             recvbuf, count, MPI_REAL_T, dst, tag, MPI_COMM_WORLD, &status);

		// copy host to device
		cuda_memcpy(GPU, Hy + offset_hy[3], recvbuf,             size_hy, cudaMemcpyHostToDevice);
		cuda_memcpy(GPU, Hz + offset_hz[3], recvbuf + length_hy, size_hz, cudaMemcpyHostToDevice);
	}

	free(sendbuf);
	free(recvbuf);
#endif
}


// PBC on +/- X boundaries : copy D2H + comm + copy H2D
void comm_cuda_pbcx(void)
{
#ifdef _MPI
	MPI_Status status;
	const int tag = 0;

	int64_t offset_hy[4], offset_hz[4], length_hy, length_hz;
	comm_buffer(offset_hy, offset_hz, &length_hy, &length_hz);

	const int count = (int)(length_hy + length_hz);
	const size_t size_hy = length_hy * sizeof(real_t);
	const size_t size_hz = length_hz * sizeof(real_t);
	real_t *sendbuf = (real_t *)malloc(size_hy + size_hz);
	real_t *recvbuf = (real_t *)malloc(size_hy + size_hz);

	// -X boundary
	if (commRank == 0) {
		// copy to buffer
		cuda_memcpy(GPU, sendbuf,             Hy + offset_hy[1], size_hy, cudaMemcpyDeviceToHost);
		cuda_memcpy(GPU, sendbuf + length_hy, Hz + offset_hz[1], size_hz, cudaMemcpyDeviceToHost);

		// MPI
		const int dst = commSize - 1;
		MPI_Sendrecv(sendbuf, count, MPI_REAL_T, dst, tag,
		             recvbuf, count, MPI_REAL_T, dst, tag, MPI_COMM_WORLD, &status);

		// copy from buffer
		cuda_memcpy(GPU, Hy + offset_hy[0], recvbuf,             size_hy, cudaMemcpyHostToDevice);
		cuda_memcpy(GPU, Hz + offset_hz[0], recvbuf + length_hy, size_hz, cudaMemcpyHostToDevice);
	}

	// +X boundary
	else if (commRank == commSize - 1) {
		// copy to buffer
		cuda_memcpy(GPU, sendbuf,             Hy + offset_hy[2], size_hy, cudaMemcpyDeviceToHost);
		cuda_memcpy(GPU, sendbuf + length_hy, Hz + offset_hz[2], size_hz, cudaMemcpyDeviceToHost);

		// MPI
		const int dst = 0;
		MPI_Sendrecv(sendbuf, count, MPI_REAL_T, dst, tag,
		             recvbuf, count, MPI_REAL_T, dst, tag, MPI_COMM_WORLD, &status);

		// copy from buffer
		cuda_memcpy(GPU, Hy + offset_hy[3], recvbuf,             size_hy, cudaMemcpyHostToDevice);
		cuda_memcpy(GPU, Hz + offset_hz[3], recvbuf + length_hy, size_hz, cudaMemcpyHostToDevice);
	}

	free(sendbuf);
	free(recvbuf);
#endif
}
