# -*- coding: utf-8 -*-
"""
OpenFDTD (Python)
Version 4.1.0
ost.py : solver
"""

import sys, time
import numpy as np
import numba
from mpi4py import MPI
import sol.input_data, sol.input_setup, sol.setup, sol.monitor
import sol.geometry, sol.plot3d_geom, sol.outputChars, sol.save_bin
import sol.solve
import sol.comm
import sol_cuda.solve

def main(argv):
    # ロゴ
    version = 'OpenFDTD (Python) Version 4.1.0'

    # 計算モード
    GPU = 1    # 0=CPU/1=GPU
    VECTOR = 0  # VECTOR: 0=OFF/1=ON

    # 入力データファイル名(ofd.pyからの相対パス)
    ofd_in = "python.ofd"
    #ofd_in = "../_tmp.ofd"
    #ofd_in = "../data/sample/1st_sample.ofd"
    #ofd_in = "../data/benchmark/benchmark100.ofd"

    # Numbaスレッド数
    thread = 8

    # 終了時のプロンプト
    prompt = False

    # 引数があるときは引数優先
    if len(argv) > 1:
        GPU, VECTOR, thread, prompt, ofd_in = _args(argv, version)
    #print(GPU, VECTOR, thread, prompt, ofd_in)

    # 型宣言(適当に変更する)
    f_dtype = 'f4'  # 'f4' or 'f8'  (単精度/倍精度)
    i_dtype = 'u1'  # 'u1' or 'i4'  (物性値種類256以下/以上)

    # MPI
    comm = MPI.COMM_WORLD
    comm_size = comm.Get_size()
    comm_rank = comm.Get_rank()

    # io : ON/OFF
    io = (comm_rank == 0)

    # 出力ファイル名
    fn_log = 'ofd.log'
    fn_out = 'ofd.npz'

    # Numbaスレッド数設定
    numba.set_num_threads(thread)

    # cpu time
    cpu = [0] * 4
    cpu[0] = _cputime(comm_size)

    # logファイルを開く
    fp_log = None
    if io:
        fp_log = open(fn_log, 'wt', encoding='utf-8')

    # 経過表示 (1)
    if io:
        # ロゴ
        logo = '<<< %s >>>\n%s, process=%d, thread=%d, vector=%s' % \
            (version, ('GPU' if GPU else 'CPU'), comm_size, thread, ('on' if VECTOR else 'off'))
        sol.monitor.log1(fp_log, logo)

    # パラメーター (辞書型 : スカラーまたは固定サイズの配列)
    Parm = sol.input_data.const()

    # パラメーター追加
    Parm['vector'] = VECTOR
    Parm['f_dtype'] = f_dtype
    Parm['i_dtype'] = i_dtype
    Parm['comm_size'] = comm_size
    Parm['comm_rank'] = comm_rank

    # [1] データ入力
    Nx = Ny = Nz = 0
    Xn = Yn = Zn = \
    iMaterial = fMaterial = iGeometry = fGeometry = \
    iFeed = fFeed = iPoint = fPoint = iInductor = fInductor = \
    Freq1 = Freq2 = None
    if io:
        Nx, Ny, Nz, Xn, Yn, Zn, \
        iMaterial, fMaterial, iGeometry, fGeometry, iFeed, fFeed, \
        iPoint, fPoint, iInductor, fInductor, Freq1, Freq2 \
        = sol.input_data.read(ofd_in, Parm)

    # 物体形状の線分を作成する (図形出力用)
    if io:
        gline, mline = sol.geometry.lines(Parm, iGeometry, fGeometry)

    # 物体形状を図形出力して終了する
    if io and (Parm['plot3dgeom'] == 1):
        sol.plot3d_geom.shape(Parm, Nx, Ny, Nz, Xn, Yn, Zn, iFeed, gline, mline)

    # (MPI) broadcast
    if Parm['comm_size'] > 1:
        Nx, Ny, Nz, Xn, Yn, Zn, \
        iMaterial, fMaterial, iGeometry, fGeometry, \
        iFeed, fFeed, iPoint, fPoint, iInductor, fInductor, Freq1, Freq2 \
        = sol.comm.broadcast(Parm,
            Nx, Ny, Nz, Xn, Yn, Zn,
            iMaterial, fMaterial, iGeometry, fGeometry,
            iFeed, fFeed, iPoint, fPoint, iInductor, fInductor, Freq1, Freq2)
    
    # セル中心
    Xc, Yc, Zc = sol.input_setup.cell_center(Xn, Yn, Zn)

    # 各種準備
    sol.input_setup.setup(
        Parm, Xn, Yn, Zn, Xc, Yc, Zc,
        iMaterial, fMaterial, iGeometry, fGeometry, iFeed, fFeed,
        iPoint, fPoint, iInductor, fInductor, Freq1, Freq2)

    # 平面波入射データ
    fPlanewave = np.zeros(15, 'f8')
    if Parm['source'] == 1:
        sol.input_setup.planewave(Parm, Xn, Yn, Zn, Freq2, fPlanewave)

    # 電磁界配列計算用の係数をParmに格納
    iMin, iMax, jMin, jMax, kMin, kMax, Ni, Nj, Nk, N0, NN \
    = sol.setup.getIndex(
        Parm['comm_size'], Parm['comm_rank'], Parm, Nx, Ny, Nz)
    #print(comm_rank, iMin, iMax, jMin, jMax, kMin, kMax, Ni, Nj, Nk, N0, NN)

    # 各種準備
    iEx, iEy, iEz, iHx, iHy, iHz, \
    C1E, C2E, C1H, C2H, \
    K1Ex, K2Ex, K1Ey, K2Ey, K1Ez, K2Ez, \
    K1Hx, K2Hx, K1Hy, K2Hy, K1Hz, K2Hz, \
    RXn, RYn, RZn, RXc, RYc, RZc, \
    fMurHx, fMurHy, fMurHz, iMurHx, iMurHy, iMurHz, \
    iPmlEx, iPmlEy, iPmlEz, iPmlHx, iPmlHy, iPmlHz, \
    gPmlXn, gPmlYn, gPmlZn, gPmlXc, gPmlYc, gPmlZc, rPmlE, rPmlH, rPml, \
    iDispEx, iDispEy, iDispEz, fDispEx, fDispEy, fDispEz, \
    cEdft, cHdft, cFdft \
    = sol.setup.setup(
        Parm, fPlanewave, Nx, Ny, Nz, Xn, Yn, Zn, Xc, Yc, Zc,
        iMaterial, fMaterial, iGeometry, fGeometry, Freq2,
        iMin, iMax, jMin, jMax, kMin, kMax, Ni, Nj, Nk, N0, NN)

    # Yee電界点を図形出力して終了する（時間がかかる）
    if io and (Parm['plot3dgeom'] == 2):
        sol.plot3d_geom.cell(Parm, Nx, Ny, Nz, Xn, Yn, Zn,
            iEx, iEy, iEz, Ni, Nj, Nk, N0)
        #ierr = 2  # 正常終了
    #_error_check(ierr, prompt)

    # 経過表示 (2)
    if io:
        sol.monitor.log2(fp_log,
            GPU, Parm, Nx, Ny, Nz, NN,
            iMaterial, iGeometry, iFeed, iPoint, Freq1, Freq2,
            iPmlEx, iPmlEy, iPmlEz, iPmlHx, iPmlHy, iPmlHz)

    cpu[1] = _cputime(comm_size)

    # FDTD反復計算ファイル名
    SOLVER = sol.solve.iteration if GPU == 0 else sol_cuda.solve.iteration
    
    # FDTD反復計算(FDTD計算の主要部)
    cEx, cEy, cEz, cHx, cHy, cHz, \
    VFeed, IFeed, VPoint, Ntime, \
    Eiter, Hiter, Iiter, Niter \
    = SOLVER(
        io, fp_log,
        Parm, fPlanewave, iFeed, fFeed, iPoint, fPoint, iInductor, fInductor, Freq2,
        Nx, Ny, Nz, Xn, Yn, Zn, Xc, Yc, Zc,
        iEx, iEy, iEz, iHx, iHy, iHz,
        C1E, C2E, C1H, C2H,
        K1Ex, K2Ex, K1Ey, K2Ey, K1Ez, K2Ez,
        K1Hx, K2Hx, K1Hy, K2Hy, K1Hz, K2Hz,
        RXn, RYn, RZn, RXc, RYc, RZc,
        fMurHx, fMurHy, fMurHz, iMurHx, iMurHy, iMurHz,
        iPmlEx, iPmlEy, iPmlEz, iPmlHx, iPmlHy, iPmlHz,
        gPmlXn, gPmlYn, gPmlZn, gPmlXc, gPmlYc, gPmlZc, rPmlE, rPmlH, rPml,
        iDispEx, fDispEx, iDispEy, fDispEy, iDispEz, fDispEz,
        cEdft, cHdft,
        iMin, iMax, jMin, jMax, kMin, kMax, Ni, Nj, Nk, N0, NN)
     
    cpu[2] = _cputime(comm_size)

    # 出力
    if io:
        # 計算結果の一部を計算しlogに出力する
        Zin, Ref, Pin, Spara, Coupling, fSurface, cSurface \
        = sol.outputChars.out(
            fp_log, Parm, Nx, Ny, Nz, Xn, Yn, Zn, Xc, Yc, Zc,
            iFeed, fFeed, iPoint, Freq1, Freq2,
            cEx, cEy, cEz, cHx, cHy, cHz,
            VFeed, IFeed, VPoint, Ntime, cFdft,
            Ni, Nj, Nk, N0, NN)

        # 経過表示 (3)
        sol.monitor.log3(fp_log, fn_log, fn_out)

        # 計算結果をファイルに出力する
        sol.save_bin.save(fn_out,
            Parm['title'], Parm['dt'], Parm['source'], fPlanewave, fFeed[:, 5],
            Ni, Nj, Nk, N0, NN,
            Xn, Yn, Zn, Freq1, Freq2,
            cEx, cEy, cEz, cHx, cHy, cHz,
            Eiter, Hiter, Iiter, Niter,
            VFeed, IFeed, VPoint, Ntime,
            Zin, Ref, Pin, Spara, Coupling,
            fSurface, cSurface, gline)

    # free
    cEx = cEy = cEz = cHx = cHy = cHz = None

    # cpu time
    cpu[3] = _cputime(comm_size)

    if io:
        # 経過表示 (4)
        sol.monitor.log4(fp_log, cpu)

        # logファイルを閉じる
        fp_log.close()

    # prompt
    if io and prompt:
        input()

# (private) CPU time
def _cputime(comm_size):
    if comm_size > 1:
        MPI.COMM_WORLD.Barrier()
        t = MPI.Wtime()
    else:
        t = time.time()
    return t

# (private) 引数処理
def _args(argv, version):

    usage = 'Usage : python ofd.py [-cpu|-gpu] [-n <thread>] [-no-vector|-vector] <datafile>'
    GPU = 0
    VECTOR = 0
    thread = 1
    prompt = 0
    ofd_in = ''

    i = 1
    while i < len(argv):
        arg = argv[i].lower()
        if   arg == '-gpu':
            GPU = 1
            i += 1
        elif arg == '-cpu':
            GPU = 0
            i += 1
        elif arg == '-vector':
            VECTOR = 1
            i += 1
        elif arg == '-no-vector':
            VECTOR = 0
            i += 1
        elif arg == '-n':
            thread = int(argv[i + 1])
            i += 2
        elif arg == '-prompt':
            prompt = 1
            i += 1
        elif arg == '--help':
            print(usage)
            sys.exit()
        elif arg == '--version':
            print(version)
            sys.exit()
        else:
            ofd_in = argv[i]
            i += 1

    return GPU, VECTOR, thread, prompt, ofd_in

# enyry point
if __name__ == "__main__":
    main(sys.argv)
