# -*- coding: utf-8 -*-
"""
comm.py (CUDA + MPI)
"""

import numpy as np
from numba import cuda
from mpi4py import MPI
import sol.comm

# X境界のHy/Hzを共有する (Sendrecv)
# こちらを採用する
def share_boundary(Parm, Hy, Hz, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']

    # buffer size
    offset_hy, offset_hz, length_hy, length_hz \
    = sol.comm.buffer_size(Parm, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0)

    # buffer
    sendbuf = np.zeros(length_hy + length_hz, Parm['f_dtype'])
    recvbuf = np.zeros(length_hy + length_hz, Parm['f_dtype'])

    ls = length_hy

    # -X境界
    if comm_rank > 0:
        # copy to buffer
        Hy[offset_hy[1]: offset_hy[1] + length_hy].copy_to_host(sendbuf[: ls])
        Hz[offset_hz[1]: offset_hz[1] + length_hz].copy_to_host(sendbuf[ls: ])

        # MPI
        dest = comm_rank - 1
        MPI.COMM_WORLD.Sendrecv(sendbuf, dest, recvbuf=recvbuf, source=dest)

        # copy from buffer
        Hy[offset_hy[0]: offset_hy[0] + length_hy] = cuda.to_device(recvbuf[: ls])
        Hz[offset_hz[0]: offset_hz[0] + length_hz] = cuda.to_device(recvbuf[ls: ])

    # +X境界
    if comm_rank < comm_size - 1:
        # copy to buffer
        Hy[offset_hy[2]: offset_hy[2] + length_hy].copy_to_host(sendbuf[: ls])
        Hz[offset_hz[2]: offset_hz[2] + length_hz].copy_to_host(sendbuf[ls: ])

        # MPI
        dest = Parm['comm_rank'] + 1
        MPI.COMM_WORLD.Sendrecv(sendbuf, dest, recvbuf=recvbuf, source=dest)

        # copy from buffer
        Hy[offset_hy[3]: offset_hy[3] + length_hy] = cuda.to_device(recvbuf[: ls])
        Hz[offset_hz[3]: offset_hz[3] + length_hz] = cuda.to_device(recvbuf[ls: ])

"""
# X境界のHy/Hzを共有する (Send + Recv)
def share_boundary(Parm, Hy, Hz, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']

    # buffer size
    offset_hy, offset_hz, length_hy, length_hz \
    = sol.comm.buffer_size(Parm, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0)

    buf = np.zeros(length_hy + length_hz, Parm['f_dtype'])
    ls = length_hy

    # +X境界送信
    if comm_rank < comm_size - 1:
        Hy[offset_hy[2]: offset_hy[2] + length_hy].copy_to_host(buf[: ls])
        Hz[offset_hz[2]: offset_hz[2] + length_hz].copy_to_host(buf[ls: :])
        MPI.COMM_WORLD.Send(buf, comm_rank + 1)

    # -X境界受信
    if comm_rank > 0:
        MPI.COMM_WORLD.Recv(buf, source = comm_rank - 1)
        Hy[offset_hy[0]: offset_hy[0] + length_hy] = cuda.to_device(buf[: ls])
        Hz[offset_hz[0]: offset_hz[0] + length_hz] = cuda.to_device(buf[ls: ])

    # -X境界送信
    if comm_rank > 0:
        Hy[offset_hy[1]: offset_hy[1] + length_hy].copy_to_host(buf[: ls])
        Hz[offset_hz[1]: offset_hz[1] + length_hz].copy_to_host(buf[ls: :])
        MPI.COMM_WORLD.Send(buf, comm_rank - 1)

    # +X境界受信
    if comm_rank < comm_size - 1:
        MPI.COMM_WORLD.Recv(buf, source = comm_rank + 1)
        Hy[offset_hy[3]: offset_hy[3] + length_hy] = cuda.to_device(buf[: ls])
        Hz[offset_hz[3]: offset_hz[3] + length_hz] = cuda.to_device(buf[ls: ])
"""

# PBC : -X/+X境界のHy/Hzを共有する (Sendrecv)
# こちらを採用する
def pbcx(Parm, Hy, Hz, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']

    # buffer size
    offset_hy, offset_hz, length_hy, length_hz \
    = sol.comm.buffer_size(Parm, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0)

    # buffer
    sendbuf = np.zeros(length_hy + length_hz, Parm['f_dtype'])
    recvbuf = np.zeros(length_hy + length_hz, Parm['f_dtype'])

    ls = length_hy

    # -X boundary
    if comm_rank == 0:
        # copy to buffer
        Hy[offset_hy[1]: offset_hy[1] + length_hy].copy_to_host(sendbuf[: ls])
        Hz[offset_hz[1]: offset_hz[1] + length_hz].copy_to_host(sendbuf[ls: ])

        # MPI
        dest = comm_size - 1
        MPI.COMM_WORLD.Sendrecv(sendbuf, dest, recvbuf=recvbuf, source=dest)

        # copy from buffer
        Hy[offset_hy[0]: offset_hy[0] + length_hy] = cuda.to_device(recvbuf[: ls])
        Hz[offset_hz[0]: offset_hz[0] + length_hz] = cuda.to_device(recvbuf[ls: ])

    # +X boundary
    elif comm_rank == comm_size - 1:
        # copy to buffer
        Hy[offset_hy[2]: offset_hy[2] + length_hy].copy_to_host(sendbuf[: ls])
        Hz[offset_hz[2]: offset_hz[2] + length_hz].copy_to_host(sendbuf[ls: ])

        # MPI
        dest = 0
        MPI.COMM_WORLD.Sendrecv(sendbuf, dest, recvbuf=recvbuf, source=dest)

        # copy from buffer
        Hy[offset_hy[3]: offset_hy[3] + length_hy] = cuda.to_device(recvbuf[: ls])
        Hz[offset_hz[3]: offset_hz[3] + length_hz] = cuda.to_device(recvbuf[ls: ])

"""
# PBC : X境界のHy/Hzをコピーする (Send + Recv)
def pbcx(Parm, Hy, Hz, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0):

    comm_size = Parm['comm_size']
    comm_rank = Parm['comm_rank']

    # buffer size
    offset_hy, offset_hz, length_hy, length_hz \
    = sol.comm.buffer_size(Parm, Ny, Nz, iMin, iMax, Ni, Nj, Nk, N0)

    buf = np.zeros(length_hy + length_hz, Parm['f_dtype'])
    ls = length_hy

    # -X境界送信
    if comm_rank == 0:
        Hy[offset_hy[1]: offset_hy[1] + length_hy].copy_to_host(buf[: ls])
        Hz[offset_hz[1]: offset_hz[1] + length_hz].copy_to_host(buf[ls: ])
        MPI.COMM_WORLD.Send(buf, comm_size - 1)

    # +X境界受信
    if comm_rank == comm_size - 1:
        MPI.COMM_WORLD.Recv(buf, source = 0)
        Hy[offset_hy[3]: offset_hy[3] + length_hy] = cuda.to_device(buf[: ls])
        Hz[offset_hz[3]: offset_hz[3] + length_hz] = cuda.to_device(buf[ls: ])

    # +X境界送信
    if comm_rank == comm_size - 1:
        Hy[offset_hy[2]: offset_hy[2] + length_hy].copy_to_host(buf[: ls])
        Hz[offset_hz[2]: offset_hz[2] + length_hz].copy_to_host(buf[ls: ])
        MPI.COMM_WORLD.Send(buf, 0)

    # -X境界受信
    if comm_rank == 0:
        MPI.COMM_WORLD.Recv(buf, source = comm_size - 1)
        Hy[offset_hy[0]: offset_hy[0] + length_hy] = cuda.to_device(buf[: ls])
        Hz[offset_hz[0]: offset_hz[0] + length_hz] = cuda.to_device(buf[ls: ])
"""
